"""Bits and pieces."""

import inspect
import os
import random
import socket
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Annotated, Callable, Generator, Generic, Iterable, NamedTuple, TypeVar

import typer

DEBUG = os.getenv("DEBUG") not in ["0", "false", "False", None]

A = TypeVar("A")
B = TypeVar("B")
T = TypeVar("T")


class Retry:
    """Retry variant."""


class Ok(NamedTuple, Generic[T]):
    """Success variant."""

    result: T


type Result[T] = Ok[T] | RuntimeError | Retry


def flatmap(f: Callable[[A], Iterable[B]], xs: Iterable[A]) -> Iterable[B]:
    """Map f over an iterable and flatten the result set."""
    return (y for x in xs for y in f(x))


def log(msg: str, stack_offset: int = 1) -> None:
    frame = inspect.stack()[stack_offset]
    module = inspect.getmodule(frame[0])
    if module is not None and getattr(module, "__file__") is not None:
        module_name = Path(getattr(module, "__file__")).with_suffix("").name
    else:
        module_name = "?"
    print(f"[{module_name}:{frame.function}]: {msg}")


def debug(msg: str) -> None:
    if DEBUG:
        log(msg, 2)


def retries[
    **P
](num_retries: int, sleep_time: float) -> Callable[
    [Callable[P, Result | Retry]], Callable[P, Result]
]:
    """ "just put a retry loop around it" """

    def wrapper(fn: Callable[P, Result | Retry]) -> Callable[P, Result]:
        def wrapped(*args: P.args, **kwargs: P.kwargs) -> Result:
            for _ in range(0, num_retries):
                match fn(*args, **kwargs):
                    case Ok(x):
                        return x
                    case Retry():
                        time.sleep(sleep_time)
                        continue
                    case RuntimeError() as e:
                        return e
            return RuntimeError(f"retries exceeded in '{fn.__name__}'")

        return wrapped

    return wrapper


@dataclass(frozen=True)
class Opts:
    """Holds common options."""

    commit_sha: str
    workflow: str
    gh_pat: str
    gh_user: str
    gh_pass: str
    gh_otp: str | None
    headless: bool
    port: int | None


def cli(main: Callable[[Opts], None]) -> Callable:
    # pylint: disable=R0913,R0917
    def _inner(
        commit_sha: Annotated[str, typer.Argument(callback=_sha_callback)],
        workflow: str,
        gh_pat: Annotated[str, typer.Option(envvar="_GH_PAT")],
        gh_user: Annotated[str, typer.Option(envvar="_GH_USER")],
        gh_pass: Annotated[str, typer.Option(envvar="_GH_PASS")],
        gh_otp: Annotated[str | None, typer.Option(envvar="_GH_OTP")] = None,
        headless: Annotated[bool, typer.Option(envvar="_HEADLESS")] = True,
        port: Annotated[int | None, typer.Option(envvar="_PORT")] = None,
    ) -> None:
        opts = Opts(
            commit_sha=commit_sha,
            workflow=workflow,
            gh_pat=gh_pat,
            gh_user=gh_user,
            gh_pass=gh_pass,
            gh_otp=gh_otp,
            headless=headless,
            port=port,
        )
        main(opts)

    return _inner


def find_free_port(min_port: int = 8100, max_port: int = 8500) -> int | None:
    num_tries = 0
    random_port = random.randint(min_port, max_port)
    while not is_port_open(random_port):
        if num_tries > 100:
            return None
        random_port = random.randint(min_port, max_port)
        num_tries += 1
    return random_port


def is_port_open(port: int) -> int:
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    res = sock.connect_ex(("127.0.0.1", port))
    sock.close()
    return res != 0


def remove_consecutive_falsy(xs: Iterable[T]) -> Generator[T, None, None]:
    yielded_empty = False
    for line in xs:
        if line:
            yielded_empty = False
            yield line
        elif not yielded_empty:
            yield line
            yielded_empty = True


def _sha_callback(value: str) -> str:
    if len(value) != 40:
        raise typer.BadParameter("need a full 40 character long commit sha")
    if value == 40 * "0":
        raise typer.BadParameter("refusing to work with the all-zero commit sha")
    return value



"""
Routines for pretty formatting the output.
"""

import multiprocessing as mp
from contextlib import suppress
from functools import partial
from queue import Empty
from typing import Callable, Dict, Generator, List

from pykka import ThreadingActor
from termcolor import colored
from termcolor._types import Color

from octotail.streamer import OutputItem
from octotail.utils import debug, flatmap, remove_consecutive_falsy

WHEEL: List[Color] = [
    "light_green",
    "light_yellow",
    "light_blue",
    "light_magenta",
    "light_cyan",
]


class Formatter(ThreadingActor):
    """The output formatting actor."""

    queue: mp.JoinableQueue
    _wheel_idx: int
    _job_map: Dict[str, int]

    def __init__(self, queue: mp.JoinableQueue):
        super().__init__()
        self.queue = queue
        self._wheel_idx = 0
        self._job_map = {}

    def _get_color(self, job_name: str) -> Color:
        if job_name == "workflow":
            return "white"

        if job_name not in self._job_map:
            self._job_map[job_name] = self._wheel_idx
            self._wheel_idx = (self._wheel_idx + 1) % len(WHEEL)
        return WHEEL[self._job_map[job_name]]

    def print_lines(self) -> None:
        while True:
            # gives us a chance to check the stop event
            with suppress(Empty):
                item = self.queue.get(timeout=2)
                self.queue.task_done()
                if item is None:
                    break
                print("\n".join(self._handle_item(item)))
        debug("exiting")

    def _handle_item(self, item: OutputItem) -> Generator[str, None, None]:
        _colored = partial(colored, color=self._get_color(item.job_name), force_color=True)
        _decorate = partial(_decorate_line, job_name=item.job_name, _colored=_colored)

        prefix = _colored(f"[{item.job_name}]:")
        return (
            f"{prefix} {line}" for line in remove_consecutive_falsy(flatmap(_decorate, item.lines))
        )


def _decorate_line(line: str, job_name: str, _colored: Callable) -> List[str]:
    if line.startswith("[command]"):
        return [colored("$ " + line.removeprefix("[command]"), "white", force_color=True)]

    if line.startswith("##[group]"):
        unprefixed = line.removeprefix("##[group]")
        rem = _colored(unprefixed, attrs=["bold"])
        sep = _colored("⎯" * (80 - len(f"remote: [{job_name}]: ") - len(unprefixed) - 6))
        return ["", f"{_colored('⎯⎯')}  {rem}  {sep}"]

    if line.startswith("##[endgroup]"):
        sep = _colored("⎯" * (80 - len(f"remote: [{job_name}]: ")))
        return [sep, ""]

    if line.startswith("##[conclusion]"):
        unprefixed = line.removeprefix("##[conclusion]").lower()
        color: Color = (
            "green" if unprefixed == "success" else "red" if unprefixed == "failure" else "yellow"
        )
        _conc_colored = partial(colored, color=color, attrs=["bold"], force_color=True)
        return ["", _conc_colored(f"Conclusion: {unprefixed.upper()}"), ""]

    return [_colored(line)]



"""GitHub actor."""

import re
from contextlib import suppress
from subprocess import CalledProcessError, check_output
from threading import Event
from typing import NamedTuple, Set

from github import Repository
from github.WorkflowRun import WorkflowRun
from pykka import ActorRef, ThreadingActor

from octotail.utils import Ok, Result, Retry, debug, log, retries

VALID_STATI = ["queued", "in_progress", "requested", "waiting", "action_required"]


class WorkflowDone(NamedTuple):
    """Workflow done message."""

    conclusion: str


class JobDone(NamedTuple):
    """Job done message."""

    job_id: int
    job_name: str
    conclusion: str


class RunWatcher(ThreadingActor):
    """Watches for changes in a GitHub Actions run."""

    wf_run: WorkflowRun
    mgr: ActorRef
    stop: Event

    _new_jobs: Set[int] = set()
    _concluded_jobs: Set[int] = set()

    def __init__(self, mgr: ActorRef, wf_run: WorkflowRun):
        super().__init__()
        self.wf_run = wf_run
        self.mgr = mgr
        self.stop = mgr.proxy().stop_event.get()

    def watch(self) -> None:
        while not self.stop.is_set():
            self.wf_run.update()

            with suppress(Exception):
                for job in self.wf_run.jobs():
                    if job.conclusion and not job.id in self._concluded_jobs:
                        self.mgr.tell(JobDone(job.id, job.name, job.conclusion))
                        self._concluded_jobs.add(job.id)
                        continue
                    if not job.id in self._new_jobs.union(self._concluded_jobs):
                        self.mgr.tell(job)
                        self._new_jobs.add(job.id)

            if self.wf_run.conclusion:
                self.mgr.tell(WorkflowDone(self.wf_run.conclusion))

            self.stop.wait(2)
        debug("exiting")


@retries(10, 0.5)
def get_active_run(
    repo: Repository, head_sha: str, workflow_name: str
) -> Result[WorkflowRun] | Retry:
    with suppress(Exception):
        runs = repo.get_workflow_runs(head_sha=head_sha)
        _runs = [r for r in runs if r.name == workflow_name]
        if not _runs:
            return Retry()

        run = _runs[0]
        if not run.status in VALID_STATI:
            log(f"cannot process run in state: '{run.status}'")
            log(f"try:\n\n\tgh run view {run.id} --log\n")
            log(f"or try browsing to:\n\n\t{run.url}\n")
            return RuntimeError("invalid run state")

        return Ok(run)

    return Retry()


def guess_repo() -> str | None:
    try:
        remotes = check_output(["git", "remote"]).decode().strip().splitlines()
    except CalledProcessError as e:
        log(f"fatal: couldn't list git remotes: {e}")
        return None
    repos = list(filter(None, map(_guess_repo, remotes)))
    if len(repos) > 1:
        log("fatal: found multiple remotes with github URLs")
    return repos[0] if repos else None


def _guess_repo(remote: str) -> str | None:
    with suppress(Exception):
        origin = check_output(f"git remote get-url {remote}".split()).decode().strip()
        if (match := re.search(r"^git@github.com:([^.]+).git$", origin)) is not None:
            return match.group(1)
    return None



"""Browser actor."""

import asyncio as aio
import json
import multiprocessing as mp
import time
from collections import deque
from contextlib import suppress
from pathlib import Path
from queue import Empty
from typing import Deque, Dict, NamedTuple, Union

from fake_useragent import UserAgent
from pykka import ActorRef, ThreadingActor
from pyppeteer import launch
from pyppeteer.browser import Browser
from pyppeteer.page import Page
from pyppeteer_stealth import stealth
from xdg.BaseDirectory import xdg_cache_home

from octotail.utils import Opts, debug, log

COOKIE_JAR = Path(xdg_cache_home) / "octotail" / "gh-cookies.json"
RANDOM_UA = UserAgent().random

CHROME_ARGS = [
    '--cryptauth-http-host ""',
    "--disable-accelerated-2d-canvas",
    "--disable-background-networking",
    "--disable-background-timer-throttling",
    "--disable-browser-side-navigation",
    "--disable-client-side-phishing-detection",
    "--disable-default-apps",
    "--disable-dev-shm-usage",
    "--disable-device-discovery-notifications",
    "--disable-extensions",
    "--disable-features=site-per-process",
    "--disable-hang-monitor",
    "--disable-java",
    "--disable-popup-blocking",
    "--disable-prompt-on-repost",
    "--disable-setuid-sandbox",
    "--disable-sync",
    "--disable-translate",
    "--disable-web-security",
    "--disable-webgl",
    "--metrics-recording-only",
    "--no-first-run",
    "--safebrowsing-disable-auto-update",
    "--no-sandbox",
    "--enable-automation",
    "--password-store=basic",
    "--use-mock-keychain",
    '--lang="en-US"',
    f'--user-agent="{RANDOM_UA}"',
]


class VisitRequest(NamedTuple):
    """Visit request message."""

    url: str
    job_id: int


class CloseRequest(NamedTuple):
    """Close request message."""

    job_id: int


class ExitRequest:
    """Exit request message."""


type BrowseRequest = Union[VisitRequest, CloseRequest, ExitRequest]


class BrowserWatcher(ThreadingActor):
    """Runs the pyppeteer browser in a separate process."""

    opts: Opts
    inbox: mp.Queue
    mgr: ActorRef

    def __init__(self, mgr: ActorRef, opts: Opts, inbox: mp.Queue):
        super().__init__()
        self.opts = opts
        self.inbox = inbox
        self.mgr = mgr

    def watch(self) -> None:
        browser = mp.Process(target=_run_browser, args=(self.opts, self.inbox))
        browser.start()
        browser.join()
        debug("exiting")
        self.mgr.stop()


def _run_browser(opts: Opts, inbox: mp.Queue) -> None:
    loop = aio.new_event_loop()
    aio.set_event_loop(loop)
    try:
        loop.run_until_complete(_browser(opts, inbox))
    except KeyboardInterrupt:
        loop.close()


async def _launch_browser(opts: Opts) -> Browser:
    return await launch(
        headless=opts.headless,
        executablePath="/usr/bin/chromium",
        options={
            "args": [*CHROME_ARGS, f"--proxy-server=127.0.0.1:{opts.port}"],
            "autoClose": False,
            "handleSIGINT": False,
        },
    )


async def _browser(opts: Opts, inbox: mp.Queue) -> None:
    browser = await _launch_browser(opts)
    tasks = set()
    open_pages: Dict[int, Page] = {}
    in_progress = aio.Event()
    visit_queue: Deque[VisitRequest] = deque()

    def _schedule_visit(_visit_req: VisitRequest) -> None:
        in_progress.set()
        _task = aio.create_task(_visit(_visit_req))
        tasks.add(_task)
        _task.add_done_callback(tasks.discard)

    async def _visit(_visit_req: VisitRequest) -> None:
        _page = await browser.newPage()
        open_pages[_visit_req.job_id] = _page
        await _page.goto(_visit_req.url, timeout=0)

    start_page = (await browser.pages())[0]
    await stealth(start_page)

    if not await _nom_cookies(opts.gh_user, start_page):
        log("logging in to GitHub")
        cookies = await _login_flow(start_page, opts)
        if isinstance(cookies, RuntimeError):
            log(f"fatal: {cookies}")
            return await browser.close()
        _save_user_cookies(opts.gh_user, cookies)

    while True:
        with suppress(Empty):
            if not in_progress.is_set() and visit_queue:
                _schedule_visit(visit_queue.pop())
                continue

            match inbox.get_nowait():
                case ExitRequest():
                    return await browser.close()

                case CloseRequest() as close_req:
                    if close_req.job_id in open_pages:
                        await open_pages[close_req.job_id].close()
                    in_progress.clear()

                case VisitRequest() as visit_req:
                    if in_progress.is_set():
                        visit_queue.appendleft(visit_req)
                    else:
                        _schedule_visit(visit_req)

        await aio.sleep(0.5)


async def _login_flow(page: Page, opts: Opts) -> list[dict] | RuntimeError:
    await page.goto("https://github.com/login")

    await page.waitForSelector("#login_field")
    await page.type("#login_field", opts.gh_user)
    await page.keyboard.press("Tab")

    await page.type("#password", opts.gh_pass)
    await page.keyboard.press("Enter")

    el = await page.waitForSelector(f"#app_totp, [data-login='{opts.gh_user}']")
    el_id = await page.evaluate("(element) => element.id", el)
    if el_id == "app_totp":
        if opts.gh_otp is None:
            return RuntimeError(
                "GitHub requested OTP authentication, but no OTP token was provided"
            )
        await el.type(opts.gh_otp)
        await page.keyboard.press("Enter")
        await page.waitForSelector(f"[data-login='{opts.gh_user}']")

    cookies = await page.cookies()
    await page.goto("about:blank")
    return cookies


async def _nom_cookies(user: str, page: Page) -> bool:
    cookies = _get_user_cookies(user)
    if cookies is None:
        return False

    if any(_is_close_to_expiry(c.get("expires", "-1")) for c in cookies):
        debug("found a stale cookie :-(")
        return False

    debug("all cookies are fresh, nom nom nom")
    await aio.gather(*[page.setCookie(c) for c in cookies])
    return True


def _save_user_cookies(user: str, cookies: list[dict]) -> None:
    COOKIE_JAR.parent.mkdir(parents=True, exist_ok=True)
    if COOKIE_JAR.exists():
        existing = json.loads(COOKIE_JAR.read_text())
    else:
        existing = {}
    COOKIE_JAR.write_text(json.dumps({**existing, **{user: cookies}}))


def _get_user_cookies(user: str) -> dict | None:
    if not COOKIE_JAR.exists():
        return None
    return json.loads(COOKIE_JAR.read_text()).get(user)


def _is_close_to_expiry(ts: str) -> bool:
    _ts, now = float(ts), time.time()
    return _ts > now and (_ts - now) < 24 * 3600
